#include "topology.h"

void hyper_bcast(int dim, int master, MPI_Comm cube,
                 void *buf, int count,
                 MPI_Datatype type, int tag,
                 MPI_Status *status)
{
	int i;
	unsigned int currank;
	unsigned int ownRank;

	MPI_Comm_rank(cube, (int*)&ownRank);

	/* follow the broadcast from the root */
	/* when it is at our "parent" node, start a recv */
	/* if we got the data, send step by step/dimension to our "children" */
	currank = master;
	for (i = 0; i < dim; i++) {
		unsigned int dimpos = 1 << i;
		unsigned int temprank = currank ^ dimpos;

		/* Someone (currank) is in the state to send us his data */
		if (temprank == ownRank) {
			MPI_Recv(buf, count, type, currank, tag, cube, status);
		}

		/* we have received our data and can send it to other nodes */
		if (currank == ownRank) {
			MPI_Send(buf, count, type, temprank, tag, cube);
		}

		/* In step i the data came a step in our direction */
		if ((temprank & dimpos) == (ownRank & dimpos)) {
			currank = temprank;
		}
	}
}
