#include "topology.h"

/*
 * Funktionen fuer die verwaltung des cartesischen topologie - cubes
 */

int cart_make_hyper(MPI_Comm srcComm, MPI_Comm *dstComm, int ndim)
{
	int *dims;
	int *periods;
	int reorder = 0;
	int i;
	int nodes;

	if (ndim != 0) {
		dims = malloc(sizeof(int) * ndim);
		periods = malloc(sizeof(int) * ndim);

		for (i = 0; i  < ndim; i++) {
			periods[i] = 0;
			dims[i] = 2;
		}
		nodes = (1 << ndim);
	} else {
		ndim = 1;
		dims = malloc(sizeof(int) * ndim);
		periods = malloc(sizeof(int) * ndim);

		for (i = 0; i  < ndim; i++) {
			periods[i] = 0;
			dims[i] = 1;
		}

		nodes = 1;
	}
	MPI_Cart_create(srcComm, ndim, dims, periods, reorder, dstComm);

	free(periods);
	free(dims);

	return 0;
}

int cart_hyper_get_neighbour(MPI_Comm cube, int maxdim, int dim, int rank, int *otherRank)
{
	int *coords;

	if (maxdim == 0) {
		(*otherRank) = 0;
	} else {
		coords = (int*)malloc(sizeof(int) * maxdim);

		MPI_Cart_coords(cube, rank, maxdim, coords);
		coords[dim] ^= 1;

		MPI_Cart_rank(cube, coords, otherRank);

		free(coords);
	}

	return 0;
}

int cart_hyper_get_dim_coord(MPI_Comm cube, int maxdim, int dim, int rank)
{

	int *tmpCoords;
	int i;

	if ((tmpCoords = (int *)calloc(maxdim, sizeof(int))) == NULL) {
		perror("cart_get_hyper_neighbour: calloc platz zum rechnen");
		return ABNORMAL_TERM;
	}
	MPI_Cart_coords(cube, rank, maxdim, tmpCoords);   /* rank -> coords */

	i = tmpCoords[ dim ] & 1;

	free(tmpCoords);

	return i;
}

void cart_hyper_global_sum(int *buffer, int len, int dim, MPI_Comm cube, int *sum)
{
	int tempsum, localsum = 0;
	int i;
	int ownRank, currank;
	int have_send = 0;
	MPI_Status status;

	MPI_Comm_rank(cube, &ownRank);

	for (i = 0; i < len; i++) {
		localsum += buffer[i];
	}

	for (i = dim - 1; i >= 0; i--) {
		cart_hyper_get_neighbour(cube, dim, i, ownRank, &currank);

		if (currank > ownRank) {
			if (have_send == 0) {
				MPI_Recv(&tempsum, 1, MPI_INT, currank, TAG, cube, &status);
				localsum += tempsum;
			}
		} else {
			if (have_send == 0) {
				MPI_Send(&localsum, 1, MPI_INT, currank, TAG, cube);
				have_send = 1;
			}
		}
	}

	hyper_bcast(dim, MASTER, cube, &localsum, 1, MPI_INT, TAG, &status);
	(*sum) = localsum;
}

void cart_hyper_global_min(int *buffer, int len, int dim, MPI_Comm cube, val_rank_pair *vrp)
{
	val_rank_pair tempmin, localmin;
	int i;
	int ownRank, currank;
	int have_send = 0;
	MPI_Status status;

	MPI_Comm_rank(cube, &ownRank);

	localmin.rank = ownRank;
	tempmin.rank = ownRank;
	do {
		tempmin.val = localmin.val;
		localmin.val <<= 1;
		localmin.val |= 1;
	} while (tempmin.val < localmin.val);
	localmin = tempmin;

	for (i = 0; i < len; i++) {
		if (localmin.val  > buffer[i]) {
			localmin.val = buffer[i];
		}
	}

	for (i = dim - 1; i >= 0; i--) {
		cart_hyper_get_neighbour(cube, dim, i, ownRank, &currank);

		if (currank > ownRank) {
			if (have_send == 0) {
				MPI_Recv(&tempmin, 2, MPI_INT, currank, TAG, cube, &status);
				if (localmin.val > tempmin.val) {
					localmin = tempmin;
				}
			}
		} else {
			if (have_send == 0) {
				MPI_Send(&localmin, 2, MPI_INT, currank, TAG, cube);
				have_send = 1;
			}
		}
	}

	hyper_bcast(dim, MASTER, cube, &localmin, 2, MPI_INT, TAG, &status);
	(*vrp) = localmin;
}
