#include "topology.h"

/*
 * Funktionen fuer die verwaltung des graphen topologie - cubes
 */

int graph_make_hyper(MPI_Comm srcComm, MPI_Comm *dstComm, int ndim)
{
	int i, j, k;
	unsigned int curnode;
	int nodes;
	int *index;
	int *edges;
	int reorder = 0;

	if (ndim != 0) {
		nodes = (1 << ndim);

		index = malloc(sizeof(int) * nodes);
		edges = malloc(sizeof(int) * nodes * ndim);

		for (i = 0; i  < nodes; i++) {
			index[i] = (i + 1) * ndim;
		}

		k = 0;
		for (i = 0; i < nodes; i++) {
			curnode = i;
			for (j = 0; j < ndim; j++) {
				edges[k] = curnode ^(1 << j);
				k++;
			}
		}
	} else {
		index = malloc(sizeof(int));
		edges = malloc(sizeof(int));

		index[0] = 0;

		nodes = (1 << ndim);
	}
	MPI_Graph_create(srcComm, nodes, index, edges, reorder, dstComm);

	free(edges);
	free(index);

	return 0;
}

int graph_hyper_get_neighbour(MPI_Comm cube, int ndim, int dim, int rank, int *otherRank)
{
	int *index;
	int *edges;
	int nodes = (1 << ndim);
	int maxedges = ndim * nodes;

	if (ndim != 0) {
		index = malloc(sizeof(int) * nodes);
		edges = malloc(sizeof(int) * maxedges);

		MPI_Graph_get(cube, nodes, maxedges, index, edges);

		(*otherRank) = edges[rank * ndim + dim];

		free(edges);
		free(index);
	} else {
		(*otherRank) = 0;
	}
	return 0;
}

void graph_hyper_global_sum(int *buffer, int len, int dim, MPI_Comm cube, int *sum)
{
	int tempsum, localsum = 0;
	int i;
	int ownRank, currank;
	int have_send = 0;
	MPI_Status status;

	MPI_Comm_rank(cube, &ownRank);

	for (i = 0; i < len; i++) {
		localsum += buffer[i];
	}

	for (i = dim - 1; i >= 0; i--) {
		graph_hyper_get_neighbour(cube, dim, i, ownRank, &currank);

		if (currank > ownRank) {
			if (have_send == 0) {
				MPI_Recv(&tempsum, 1, MPI_INT, currank, TAG, cube, &status);
				localsum += tempsum;
			}
		} else {
			if (have_send == 0) {
				MPI_Send(&localsum, 1, MPI_INT, currank, TAG, cube);
				have_send = 1;
			}
		}
	}

	hyper_bcast(dim, MASTER, cube, &localsum, 1, MPI_INT, TAG, &status);
	(*sum) = localsum;
}

void graph_hyper_global_min(int *buffer, int len, int dim, MPI_Comm cube, val_rank_pair *vrp)
{
	val_rank_pair tempmin, localmin;
	int i;
	int ownRank, currank;
	int have_send = 0;
	MPI_Status status;

	MPI_Comm_rank(cube, &ownRank);

	localmin.rank = ownRank;
	tempmin.rank = ownRank;
	do {
		tempmin.val = localmin.val;
		localmin.val <<= 1;
		localmin.val |= 1;
	} while (tempmin.val < localmin.val);
	localmin = tempmin;

	for (i = 0; i < len; i++) {
		if (localmin.val  > buffer[i]) {
			localmin.val = buffer[i];
		}
	}

	for (i = dim - 1; i >= 0; i--) {
		graph_hyper_get_neighbour(cube, dim, i, ownRank, &currank);

		if (currank > ownRank) {
			if (have_send == 0) {
				MPI_Recv(&tempmin, 2, MPI_INT, currank, TAG, cube, &status);
				if (localmin.val > tempmin.val) {
					localmin = tempmin;
				}
			}
		} else {
			if (have_send == 0) {
				MPI_Send(&localmin, 2, MPI_INT, currank, TAG, cube);
				have_send = 1;
			}
		}
	}

	hyper_bcast(dim, MASTER, cube, &localmin, 2, MPI_INT, TAG, &status);
	(*vrp) = localmin;
}
