#include "topology.h"

/* hauptprogramm */
int main(int argc, char** argv)
{

	int auswahl;      /* die auswahl aus dem menu */
	int done;       /* abbruchflag */

	MPI_Status status;     /* status von receive operationen */

	MPI_Comm  hyper_comm_own;   /* Communicator fuer eigenen Hypercube */
	MPI_Comm  hyper_comm_cart;   /* Communicator fuer cartesischen Hypercube */
	MPI_Comm  hyper_comm_graph;   /* Communicator fuer graph Hypercube */

	int me, global_size;    /* mein rang und die ganze groesse */
	int cube_dim;      /* dimension der hypercubes */
	int cube_size;      /* anzahl der knoten in den hypercube */

	int random_data_count;    /* anzahl der zufaellig erzeugten daten */
	int *random_data;     /* array zufaellig erzeugter daten */

	int sum;       /* globale summe, die mit den eigenen */
	/* Funktionen berechnet wurde */
	int ref_sum;      /* globale summe, die mit den MPI-Funktionen */
	/* berechnet wurde */
	val_rank_pair minpair;    /* globales minimum + rank, das mit eigenen */
	/* Funktionen berechnet wurde */
	val_rank_pair ref_minpair;   /* globales minimum + rank, das mit MPI-
 Funktionen berechnet wurde */

	/*
	 * 1. MPI-Initialisierung
	 */
	MPI_Init(&argc, &argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &me);
	MPI_Comm_size(MPI_COMM_WORLD, &global_size);

	/*
	 * 2. erzeugen der 3 hypercubes
	 */
	if (make_hyper(MPI_COMM_WORLD, &hyper_comm_own, &cube_dim, &cube_size) != 0) {
		return 1;
	}

	if (me < cube_size) {
		if (graph_make_hyper(hyper_comm_own, &hyper_comm_graph, cube_dim) != 0) {
			return 1;
		}

		if (cart_make_hyper(hyper_comm_own, &hyper_comm_cart, cube_dim) != 0) {
			return 1;
		}
	}


	if (me == MASTER) {
		printf("Gesamte Anzahl der Knoten     : %d\n", global_size);
		printf("Dimmension des Hypercubes     : %d\n", cube_dim);
		printf("Anzahl der Knoten im Hypercube: %d\n\n", cube_size);
	}

	/*
	 da nach erzeugen des hypercubes evtl. ein paar prozesse sich nicht in
	 diesen befunden, muessen sie von den weiteren aktivitaeten ausgeschlossen
	 werden.
	 das kriterium nachdem dies geschieht, ist, wenn ihr rank groesser als
	 der groesste im maximalen hypercube fassbare rank ist.
	*/

	if (me < cube_size) {
		done = FALSE;
		while (!done) {
			MPI_Barrier(hyper_comm_own);

			if (me == MASTER) {
				auswahl = 0;
				/* der master stellt das menue dar. */
				printf("Hauptmenue\n\n");
				printf("1.\tGlobale Summe auf Cartesischer Topoloy\n");
				printf("2.\tGlobales Minimum auf Cartesischer Topoloy\n");
				printf("3.\tGlobale Summe auf Graph-Topoloy\n");
				printf("4.\tGlobales Minimum auf Graph-Topoloy\n");
				printf("5.\tVgl: Cart. Top <-> MPI-Funktionen\n");
				printf("6.\tVgl: Graph. Top <-> MPI-Funktionen\n");
				printf("7.\tExit\n");
				printf("\nAuswahl: ");
				fflush(stdout);
				scanf("%d", &auswahl);
			}

			/*
			* 3. auswahl ergebnis bekanntgeben mittels des eigenen hypercube-broadcasts
			*/
			hyper_bcast(cube_dim, MASTER, hyper_comm_own, &auswahl, 1, MPI_INT, TAG, &status);

#ifdef DEBUG
			printf("%d: auswahl %d bekommen!\n", me, auswahl);
			fflush(stdout);
#endif

			switch (auswahl) {
			case 1:
				if (me == MASTER)
					printf("Summe bei Kartesischer Topologie\n");

				/*
				* 4. daten erzeugen und verteilen.
				*    wenn die daten mittels hyper_init_data an alle prozesse geshickt wurden,
				*    also jeder knoten des eigenen hypercubes die daten hat, dann hat sie auch
				*    jeder knoten im kartesischen und Graph-Hypercube.
				*/
				hyper_init_data(MASTER, MPI_COMM_WORLD, cube_dim, TAG, hyper_comm_own, &status, &random_data_count, &random_data);

#ifdef OUTPUT_VAL
				printf("%d: %d daten empfangen!\n", me, random_data_count);
				for (i = 0; i < random_data_count; i++) {
					printf("%d ", random_data[ i ]);
				}
				printf("\n");
				fflush(stdout);
#endif

				/*
				* 5. berechnen der globalen summe mittels cart_hyper_global_sum
				*/
				cart_hyper_global_sum(random_data, random_data_count, cube_dim, hyper_comm_cart, &sum);

				if (me == MASTER) {
					printf("Summe aller Elemente: %d\n", sum);
					fflush(stdout);
				}
				break;

			case 2:
				if (me == MASTER)
					printf("Minimum bei Kartesischer Topologie\n");

				/*
				* 6. daten erzeugen und verteilen (siehe 4.)
				*/
				hyper_init_data(MASTER, MPI_COMM_WORLD, cube_dim, TAG, hyper_comm_own, &status, &random_data_count, &random_data);

#ifdef OUTPUT_VAL
				printf("%d: %d daten empfangen!\n", me, random_data_count);
				for (i = 0; i < random_data_count; i++) {
					printf("%d ", random_data[ i ]);
				}
				printf("\n");
				fflush(stdout);
#endif

				/*
				* 7. berechnen des globalen minimums mittels cart_hyper_global_min
				*/
				cart_hyper_global_min(random_data, random_data_count, cube_dim, hyper_comm_cart, &minpair);

				if (me == MASTER) {
					printf("Minimum aller Elemente: %d bei Prozess %d\n", minpair.val, minpair.rank);
					fflush(stdout);
				}
				break;

			case 3:
				if (me == MASTER)
					printf("Summe bei Graph Topologie\n");

				/*
				* 8. daten erzeugen und verteilen (siehe 4.)
				*/
				hyper_init_data(MASTER, MPI_COMM_WORLD, cube_dim, TAG, hyper_comm_own, &status, &random_data_count, &random_data);

#ifdef OUTPUT_VAL
				printf("%d: %d daten empfangen!\n", me, random_data_count);
				for (i = 0; i < random_data_count; i++) {
					printf("%d ", random_data[ i ]);
				}
				printf("\n");
				fflush(stdout);
#endif

				/*
				* 9. berechnen der globalen summe mittels graph_hyper_global_sum
				*/
				graph_hyper_global_sum(random_data, random_data_count, cube_dim, hyper_comm_graph, &sum);

				if (me == MASTER) {
					printf("Summe aller Elemente: %d\n", sum);
					fflush(stdout);
				}
				break;

			case 4:
				if (me == MASTER)
					printf("Minimum bei Graph Topologie\n");

				/*
				* 10. daten erzeugen und verteilen (siehe 4.)
				*/
				hyper_init_data(MASTER, MPI_COMM_WORLD, cube_dim, TAG, hyper_comm_own, &status, &random_data_count, &random_data);

#ifdef OUTPUT_VAL
				printf("%d: %d daten empfangen!\n", me, random_data_count);
				for (i = 0; i < random_data_count; i++) {
					printf("%d ", random_data[ i ]);
				}
				printf("\n");
				fflush(stdout);
#endif

				/*
				* 11. berechnen des globalen minimums mittels graph_hyper_global_min
				*/
				graph_hyper_global_min(random_data, random_data_count, cube_dim, hyper_comm_graph, &minpair);

				if (me == MASTER) {
					printf("Minimum aller Elemente: %d bei Prozess %d\n", minpair.val, minpair.rank);
					fflush(stdout);
				}
				break;

			case 5:
				if (me == MASTER)
					printf("Funktionsvergleich Kart. Funktionen <-> MPI-Funktionen\n");

				/*
				* 12. daten erzeugen und verteilen (siehe 4.)
				*/
				hyper_init_data(MASTER, MPI_COMM_WORLD, cube_dim, TAG, hyper_comm_own, &status, &random_data_count, &random_data);

#ifdef OUTPUT_VAL
				printf("%d: ich soll %d daten empfangen!\n", me, random_data_count);
				for (i = 0; i < random_data_count; i++) {
					printf("%d ", random_data[ i ]);
				}
				printf("\n");
				fflush(stdout);
#endif

				/*
				* 13. berechnen der globalen summe mittels cart_hyper_global_sum und hyper_global_sum
				*     und ausgabe des vergleichs-ergebnisses
				*/
				cart_hyper_global_sum(random_data, random_data_count, cube_dim, hyper_comm_cart, &sum);
				hyper_global_sum(random_data, random_data_count, cube_dim, hyper_comm_cart, &ref_sum);

				if (me == MASTER) {
					if (sum == ref_sum) {
						printf("Vergleich erfolgreich.\nBeide Summenfunktionen liefern: %d\n", sum);
					} else {
						printf("Vergleich nicht erfolgreich.\n");
						printf("Kart. Summenfunktionen liefert: %d\n", sum);
						printf("MPI   Summenfunktionen liefert: %d\n", ref_sum);
					}
					fflush(stdout);
				}

				/*
				* 14. berechnen des globalen minimums mittels cart_hyper_global_min und hyper_global_min
				*     und ausgabe des vergleichs-ergebnisses
				*/
				cart_hyper_global_min(random_data, random_data_count, cube_dim, hyper_comm_cart, &minpair);
				hyper_global_min(random_data, random_data_count, cube_dim, hyper_comm_cart, &ref_minpair);

				if (me == MASTER) {
					if (minpair.val == ref_minpair.val) {
						printf("Vergleich erfolgreich.\nBeide Minimumfunktionen liefern: %d\n", minpair.val);
						if (minpair.rank == ref_minpair.rank) {
							printf("Heimatprozess des Minimums ist eindeutig: %d\n", minpair.rank);
						} else {
							printf("Heimatprozesse des Minimums sind: %d und %d\n", minpair.rank, ref_minpair.rank);
						}
					} else {
						printf("Vergleich nicht erfolgreich.\n");
						printf("Kart. Minimumfunktionen liefert: %d\n", minpair.val);
						printf("MPI   Minimumfunktionen liefert: %d\n", ref_minpair.val);
					}
					fflush(stdout);
				}
				break;

			case 6:
				if (me == MASTER)
					printf("Funktionsvergleich Graph Funktionen <-> MPI-Funktionen\n");

				/*
				* 15. daten erzeugen und verteilen (siehe 4.)
				*/
				hyper_init_data(MASTER, MPI_COMM_WORLD, cube_dim, TAG, hyper_comm_own, &status, &random_data_count, &random_data);

#ifdef OUTPUT_VAL
				printf("%d: ich soll %d daten empfangen!\n", me, random_data_count);
				for (i = 0; i < random_data_count; i++) {
					printf("%d ", random_data[ i ]);
				}
				printf("\n");
				fflush(stdout);
#endif

				/*
				* 16. berechnen der globalen summe mittels graph_hyper_global_sum und hyper_global_sum
				*     und ausgabe des vergleichs-ergebnisses
				*/
				graph_hyper_global_sum(random_data, random_data_count, cube_dim, hyper_comm_graph, &sum);
				hyper_global_sum(random_data, random_data_count, cube_dim, hyper_comm_graph, &ref_sum);

				if (me == MASTER) {
					if (sum == ref_sum) {
						printf("Vergleich erfolgreich.\nBeide Summenfunktionen liefern: %d\n", sum);
					} else {
						printf("Vergleich nicht erfolgreich.\n");
						printf("Graph Summenfunktionen liefert: %d\n", sum);
						printf("MPI   Summenfunktionen liefert: %d\n", ref_sum);
					}
					fflush(stdout);
				}

				/*
				* 17. berechnen des globalen minimums mittels cart_hyper_global_min und hyper_global_min
				*     und ausgabe des vergleichs-ergebnisses
				*/
				graph_hyper_global_min(random_data, random_data_count, cube_dim, hyper_comm_graph, &minpair);
				hyper_global_min(random_data, random_data_count, cube_dim, hyper_comm_graph, &ref_minpair);

				if (me == MASTER) {
					if (minpair.val == ref_minpair.val) {
						printf("Vergleich erfolgreich.\nBeide Minimumfunktionen liefern: %d\n", minpair.val);
						if (minpair.rank == ref_minpair.rank) {
							printf("Heimatprozess des Minimums ist eindeutig: %d\n", minpair.rank);
						} else {
							printf("Heimatprozesse des Minimums sind: %d und %d\n", minpair.rank, ref_minpair.rank);
						}
					} else {
						printf("Vergleich nicht erfolgreich.\n");
						printf("Graph Minimumfunktionen liefert: %d\n", minpair.val);
						printf("MPI   Minimumfunktionen liefert: %d\n", ref_minpair.val);
					}
					fflush(stdout);
				}
				break;
			case 7:
				done = TRUE;
				break;
			default:
				if (me == MASTER) {
					printf("Naja, eine Ziffer zwischen 1 und 7 war das aber nicht!\n");
					fflush(stdout);
				}
			} /* switch( ... */

		} /* while( !done )... */

	} /* if ( me < cube_size )... */


	/*
	 * 18. MPI beenden
	 */
	MPI_Comm_free(&hyper_comm_own);
	MPI_Comm_free(&hyper_comm_cart);
	MPI_Comm_free(&hyper_comm_graph);
	MPI_Finalize();

	return 0;

}
