#include "topology.h"

int make_hyper(MPI_Comm srcComm, MPI_Comm *dstComm, int *dim, int *size)
{
	unsigned int oldsize;
	MPI_Group world_group, new_group;
	int ranges[1][3];

	(*dim) = 0;
	MPI_Comm_size(srcComm, (int*)&oldsize);

	while (oldsize > 0) {
		(*dim)++;
		oldsize >>= 1;
	}
	(*dim)--;

	if ((*dim) > 0) {
		(*size) = 1 << (*dim);
	} else {
		(*dim) = 0;
		(*size) = 1;
	}

	MPI_Comm_group(srcComm, &world_group);
	ranges[0][0] = 0; /* erster eintrag */
	ranges[0][1] = (*size) - 1; /* letzter eintrag */
	ranges[0][2] = 1; /* schrittweite */
	MPI_Group_range_incl(world_group, 1, ranges, &new_group);
	MPI_Comm_create(srcComm, new_group, dstComm);

	MPI_Group_free(&world_group);
	MPI_Group_free(&new_group);

	return 0;
}

int hyper_init_data(int master_in_world, MPI_Comm world_comm,
                    int dim, int tag,
                    MPI_Comm cube, MPI_Status *status,
                    int *count, int **data)
{

	int ownRank;
	int local_ownRank;
	int anzahl;
	int eingabe;
	int i;
	unsigned int size = 1 << dim;
	int rest;

	MPI_Comm_rank(world_comm, &ownRank);
	MPI_Comm_rank(cube, &local_ownRank);

	if (ownRank == master_in_world) {
		printf("\nAnzahl Elemente: ");
		fflush(stdout);
		scanf("%d", &anzahl);
	}

	hyper_bcast(dim, 0, cube, &anzahl, 1, MPI_INT, tag, status);
	(*count) = anzahl / size;
	rest = anzahl % size;

	if (local_ownRank < rest) {
		(*count)++;
	}

	(*data) = (int*)malloc(sizeof(int) * (*count));

	if (ownRank == master_in_world) {
		for (i = 0; i < anzahl; i++) {
			printf("Zahl %d:", i);
			fflush(stdout);
			scanf("%d", &eingabe);

			if ((i % size) == 0) {
				(*data)[i / size] = eingabe;
			} else {
				MPI_Send(&eingabe, 1, MPI_INT, i % size, tag, cube);
			}
		}
	} else {
		for (i = 0; i < (*count); i++) {
			MPI_Recv(&eingabe, 1, MPI_INT, MASTER, tag, cube, status);
			(*data)[i] = eingabe;
		}
	}

	return 0;

}

void hyper_global_sum(int *buffer, int len, int dim, MPI_Comm cube, int *sum)
{
	int tempsum, localsum = 0;
	int i;
	unsigned int ownRank;
	MPI_Status status;

	MPI_Comm_rank(cube, (int*)&ownRank);

	for (i = 0; i < len; i++) {
		localsum += buffer[i];
	}

	MPI_Reduce(&localsum, &tempsum, 1, MPI_INT, MPI_SUM, MASTER, cube);

	hyper_bcast(dim, MASTER, cube, &tempsum, 1, MPI_INT, TAG, &status);
	(*sum) = tempsum;
}

void hyper_global_min(int *buffer, int len, int dim, MPI_Comm cube, val_rank_pair *vrp)
{
	val_rank_pair tempmin, localmin;
	int i;
	unsigned int ownRank;
	MPI_Status status;

	MPI_Comm_rank(cube, (int*)&ownRank);

	localmin.rank = ownRank;
	tempmin.rank = ownRank;
	do {
		tempmin.val = localmin.val;
		localmin.val <<= 1;
		localmin.val |= 1;
	} while (tempmin.val < localmin.val);
	localmin = tempmin;

	for (i = 0; i < len; i++) {
		if (localmin.val  > buffer[i]) {
			localmin.val = buffer[i];
		}
	}

	MPI_Reduce(&localmin, &tempmin, 1, MPI_2INT, MPI_MINLOC, MASTER, cube);

	hyper_bcast(dim, MASTER, cube, &tempmin, 2, MPI_INT, TAG, &status);
	(*vrp) = tempmin;
}
